#include <GL/glut.h>
#include "marching_cube.h"
#include <string.h>
#include <stdio.h>
#include <assert.h>
#include "Time_Counter.h"
#include <iostream>
using namespace std;

#define MATRIX_SIZE 448
#define STEP 16
#define STEP_B 4
#define SPEED 10

class Metaballs3D;

#define nb_metaballs 20
Metaballs3D * tab_meta[nb_metaballs];
double *texture_map;


bool DISPLAY_GRID = 0;
bool ROTATE = 0;
double  SEUIL = 6.0;
bool DISPLAY_BOX;

int PX_M2 = 335;


class Spheric_Potential;
void Display_Potential_Field_Bounding_Box( Potential_Field * field);

class Spheric_Potential : public Potential_Field
{
public :
	Spheric_Potential(int size , int radius):Potential_Field(size)
	{
		this->Radius = radius;
	};
	double Get_Potential( Point3D p)
	{
		int size = Get_Size();
		if( p.x == size/2 && p.y == size/2 && p.z == size/2 )
			return 1000.0;
		return 1000.0/((p.x - size/2)*(p.x - size/2)+(p.y - size/2)*(p.y - size/2)+(p.z - size/2)*(p.z - size/2));
	};
private :
	int Radius;
};



class Potential_Matrix : public Potential_Field
{
public :
	void Reset()
	{		
		int l = ((Size>>STEP_B)+1)*((Size>>STEP_B)+1)*((Size>>STEP_B)+1)*sizeof(double);		
		memset(matrix , 0 , l);
	};

	Potential_Matrix(int size):Potential_Field(size)
	{
		cout << "contructeur de matrix , taille du monde : " << size << "  cases:" << STEP<<"*"<<STEP<<"*"<<STEP;
		int l = (size>>STEP_B)+1;
		cout << "   largeur de la matrice : " << l <<"   ";		
		matrix = new double[l*l*l];
		if( matrix == NULL)
			cout << "error : not enough memory !!! \n" ;
		this->Reset();
		printf(" [ OK ] \n");
	};
	inline double Get_Potential( Point3D p)
	{		
		return matrix[(p.x>>STEP_B) + (p.y>>STEP_B)*((Size>>STEP_B)+1) + (p.z>>STEP_B)*((Size>>STEP_B)+1)*((Size>>STEP_B)+1)];
	};
	inline void Set_Potential_world( int x ,int y, int z , double p)
	{		
		matrix[(x>>STEP_B) + (y>>STEP_B)*((Size>>STEP_B)+1) + (z>>STEP_B)*((Size>>STEP_B)+1)*((Size>>STEP_B)+1)]+=p;
	};
	inline void Set_Potential_matrix( int x ,int y, int z , double p)
	{	
		matrix[x + y*((Size>>STEP_B)+1) + z*((Size>>STEP_B)+1)*((Size>>STEP_B)+1)]+=p;
	};
	void print_line_yz(int y,int z)
	{
		cout << "Size : " << Size <<"\n";
		cout << "STEP_B : " << STEP_B <<"\n";
		cout << "Size>>STEp_B : " << (Size>>STEP_B)<< "\n";
		for(int x=0;x<Size;x+=STEP)
			cout << "val en ["<<x<<","<<y<<","<<z<<"] = "<< Get_Potential( Point3D(x,y,z)) << "\n";
	};

private :
	double *matrix;
};





void Init_Texture()
{
  printf("texture a initialiser ! \n");

  texture_map = new double[257*257*257];
  if( texture_map == NULL)
  {
  	cout << "not enough memory for the 3D texture !! \n";
	exit(0);
  }

  memset( texture_map , 0 , 257*257*257*sizeof(double));

  double potential;
  for(int x=-128 ; x <= 128 ; x++)
     for(int y=-128 ; y <= 128 ; y++)
     	for(int z=-128 ; z <= 128 ; z++)
		if( x*x+y*y+z*z < 128*128)
		{
			if( x==0 && y==0 && z==0)
				potential = 16384.0;
			else
				potential = ((16384.0 / ( x*x + y*y + z*z))-1.0);
				
			texture_map[(x+128)+(y+128)*257+(z+128)*257*257] = potential;			
		}
		else
			texture_map[(x+128)+(y+128)*257+(z+128)*257*257] = 0.0;
			

  printf("texture initialise ! \n");
}

void Display_Cube(int x,int y,int z,int l)
{
	glBegin( GL_LINES);
		glVertex3i(x,y,z);
		glVertex3i(x+l,y,z);
		glVertex3i(x+l,y,z);
		glVertex3i(x+l,y+l,z);
		glVertex3i(x+l,y+l,z);
		glVertex3i(x,y+l,z);
		glVertex3i(x,y+l,z);
		glVertex3i(x,y,z);

		glVertex3i(x,y,z+l);
		glVertex3i(x+l,y,z+l);
		glVertex3i(x+l,y,z+l);
		glVertex3i(x+l,y+l,z+l);
		glVertex3i(x+l,y+l,z+l);
		glVertex3i(x,y+l,z+l);
		glVertex3i(x,y+l,z+l);
		glVertex3i(x,y,z+l);

		glVertex3i(x,y,z);
		glVertex3i(x,y,z+l);

		glVertex3i(x+l,y,z);
		glVertex3i(x+l,y,z+l);

		glVertex3i(x+l,y+l,z);
		glVertex3i(x+l,y+l,z+l);

		glVertex3i(x,y+l,z);
		glVertex3i(x,y+l,z+l);

	glEnd();
}

class Metaballs3D
{
public :
	Metaballs3D(int px , int py , int pz, int rayon)
	{
			Px=px;
			Py=py;
			Pz=pz;
			Rayon=rayon;
			Vx=10;
			Vy=10;
			Vz=10;
			cout << "new metaball 3D en " << Px << ","<< Py << ","<< Py << "   de rayon:" << Rayon << "\n";

	}
	void Display_Bounding_Box()
	{
		Display_Cube(Px-Rayon , Py-Rayon , Pz-Rayon , 2*Rayon+1);
	}
	void AddToMatrix( Potential_Matrix * world )
	{

		unsigned char r =  (unsigned char) Rayon;
		unsigned char rapport=7;

		while( !(r & 0x01 ) )
		{
			r=r>>1;
			rapport--;
		}

		// si rayon = 128 --> rapport=0
		// si rayon = 64  --> rapport = 1

		//cout << "Px="<< Px << "  Py=" << Py << "   rayon =" << Rayon <<"\n";
		//cout << "rapport=" << (int)rapport << "\n";
		
		int x0 = (Px - Rayon) >> STEP_B;
		int offsetx = (x0<<STEP_B) - (Px - Rayon);
		if( offsetx != 0)
		{
			x0++;
			offsetx+=STEP;
		}



		
		int x1 = ( Px +Rayon )>>STEP_B;		
		int y0 = (Py - Rayon) >> STEP_B;
		
		int offsety = (y0<<STEP_B) - (Py - Rayon);
		if( offsety != 0)
		{
			y0++;
			offsety=STEP;
		}
		
		int y1 = ( Py +Rayon )>>STEP_B;	
		int z0 = (Pz - Rayon) >> STEP_B;

		
		int offsetz = (z0<<STEP_B) - (Pz - Rayon);
		if( offsetz != 0)
		{
			z0++;
			offsetz+=STEP;
		}

		
		int z1 = ( Pz +Rayon )>>STEP_B;		

		//cout << x0 << "," << x1 << "...." << y0 << "," << y1 << "......" << z0 << "," << z1 << "\n";

		double potential;
		/*
		cout << "box->x'x :" << x0 << "," << x1 << "\n";
		cout << "offset x: " << offsetx << "\n";
		cout << "rapport :" << (int)rapport << "\n";
		*/
		for(int x=x0 ; x<=x1;x++)
			for(int y=y0 ; y<=y1;y++)
				for(int z=z0 ; z<=z1;z++)
				if(    (x << STEP_B ) >=0 &&  (x << STEP_B ) <= MATRIX_SIZE
				    && (y << STEP_B ) >=0 &&  (y << STEP_B ) <= MATRIX_SIZE
				    && (z << STEP_B ) >=0 &&  (z << STEP_B ) <= MATRIX_SIZE )
				{					
					potential = texture_map[ ( (offsetx+((x-x0)<<STEP_B))<<rapport)
								+(((offsety+((y-y0)<<STEP_B))<<rapport)*257)
								+(((offsetz+((z-z0)<<STEP_B))<<rapport)*257*257)];
					world->Set_Potential_matrix( x , y , z , potential);					
				}




	}
	
	void Move( )
	{
		if( (Px + Vx) < Rayon || (Px+Vx) >=MATRIX_SIZE-Rayon || (Py + Vy) < Rayon || (Py+Vy) >=MATRIX_SIZE-Rayon
		  || (Pz + Vz) < Rayon || (Pz+Vz) >=MATRIX_SIZE-Rayon)
		{
			do{
				Vx=rand()%(SPEED*2)-SPEED;
				Vy=rand()%(SPEED*2)-SPEED;
				Vz=rand()%(SPEED*2)-SPEED;
			}while( Vx==0 && Vy==0 && Vz==0);

		}
		else
		{
			Px += Vx;
			Py += Vy;
			Pz += Vz;
		}
	}
	inline int Get_Px() { return Px;};
	inline int Get_Py() { return Py;};
	inline int Get_Pz() { return Pz;};
	inline void Set_Px(int px) { Px=px;};
private :
	int Px,Py,Pz,Rayon;
	int Vx,Vy,Vz;
};



Potential_Matrix * The_Matrix;
int rot=0;

void init(void)
{
	Init_Texture();
	glClearColor( 0.0 ,  0.0 , 0.0 , 0.0 );
	glShadeModel(GL_SMOOTH);


	GLfloat light_pos[] = { 1.0 , 1.0 , -1.0 , 0.0};
	GLfloat white_light[] = { 1.0 , 1.0 , 1.0 , 1.0};
	GLfloat mat_specular[] = { 1.0 , 1.0 , 1.0 , 1.0};
	GLfloat ambiant[] = { 1.0 , 0.0 , 0.0 , 1.0};
	GLfloat red[] = { 1.0 , 0.0 , 0.0 , 1.0};

	GLfloat mat_shininess[] = { 1.0};

	glMaterialfv( GL_FRONT_AND_BACK , GL_DIFFUSE , red);
	glMaterialfv( GL_FRONT_AND_BACK , GL_SPECULAR , mat_specular);
	glMaterialfv( GL_FRONT_AND_BACK , GL_SHININESS , mat_shininess);

	glLightfv( GL_LIGHT0 , GL_AMBIENT , ambiant );
	glLightfv( GL_LIGHT0 , GL_POSITION , light_pos );
	glLightfv( GL_LIGHT0 , GL_DIFFUSE , white_light );
	glLightfv( GL_LIGHT0 , GL_SPECULAR , white_light );

	glEnable( GL_LIGHT0);
	glEnable(GL_DEPTH_TEST);
	
	The_Matrix = new Potential_Matrix( MATRIX_SIZE );
	for(int i=0;i<nb_metaballs;i++)
  		tab_meta[i] = new Metaballs3D( 200,200,200,128);//64<<(rand()%2));
}


void Display_Axis()
{
	glBegin( GL_LINES);
		glColor3f(500.0,0.0,0.0);
		glVertex3f(0.0,0.0,0.0);
		glVertex3f(500.0,0.0,0.0);

		glColor3f(0.0,500.0,0.0);
		glVertex3f(0.0,0.0,0.0);
		glVertex3f(0.0,500.0,0.0);

		glColor3f(0.0,0.0,500.0);
		glVertex3f(0.0,0.0,0.0);
		glVertex3f(0.0,0.0,500.0);
	glEnd();
}

void Display_Grid()
{
	int x,y,z;

	glBegin( GL_LINES);
	glColor3f(0.0,0.7,0.0);
	for( x=0;x<=MATRIX_SIZE;x+=STEP)
		for( y=0;y<=MATRIX_SIZE;y+=STEP)
		{
			glVertex3i(x,y,0);
			glVertex3i(x,y,MATRIX_SIZE);
		}
	for( x=0;x<=MATRIX_SIZE;x+=STEP)
		for( z=0;z<=MATRIX_SIZE;z+=STEP)
		{
			glVertex3i(x,0,z);
			glVertex3i(x,MATRIX_SIZE,z);
		}
	for( z=0;z<=MATRIX_SIZE;z+=STEP)
		for( y=0;y<=MATRIX_SIZE;y+=STEP)
		{
			glVertex3i(0,y,z);
			glVertex3i(MATRIX_SIZE,y,z);
		}


	glEnd();

}

void display(void)
{

	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	glColor3f( 1.0 , 0.0 , 0.0 );
	glLoadIdentity();
	gluLookAt( 0.0 ,0.0 , -600.0 , 0.0 , 0.0, 0.0, 0.0, 1.0, 0.0 );


	glRotatef( (float) rot , 0.0 , 1.0 , 0.0);
	glTranslatef( -160.0 , -160.0 , -160.0 );
	glDisable(GL_LIGHTING);
	if( DISPLAY_GRID )
		Display_Grid();
	if( DISPLAY_BOX)
	{
		Display_Axis();
		Display_Potential_Field_Bounding_Box(The_Matrix );		
	}
	glEnable(GL_LIGHTING);



	
	Marching_Cube_Begin(The_Matrix,STEP);
	for(int i=0;i<nb_metaballs;i++)
		Draw_Iso_Surface_Around_Point( The_Matrix , SEUIL , STEP ,
				(tab_meta[i]->Get_Px()/STEP)*STEP,
				(tab_meta[i]->Get_Py()/STEP)*STEP,
				(tab_meta[i]->Get_Pz()/STEP)*STEP);	
	Marching_Cube_End();
	glFlush();
	glutSwapBuffers(); // In double buffered mode we invert the positions of the visible buffer and the writing buffer



}

void idle(void)
{
	if( ROTATE) { rot+=5; if( rot >=360) rot = 0; }
	Time_Counter counter;

	The_Matrix->Reset();
	counter.restart();
	for(int i=0;i<nb_metaballs;i++)
	{
		tab_meta[i]->Move();
		tab_meta[i]->AddToMatrix(The_Matrix);
	}

	counter.stop();
	//cout << "ajout des metaballs ds la matrice :" ; counter.print();
	glutPostRedisplay();

}

void Display_Potential_Field_Bounding_Box( Potential_Field * field)
{
	glColor3f(0.7,0.0,0.0);
	Display_Cube(0,0,0,field->Get_Size());

}

void reshape(int w, int h)
{
	glViewport( 0 , 0 , (GLsizei) w , (GLsizei) h );
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective(60.0,640/480,1,100000);
	glMatrixMode( GL_MODELVIEW);
}


void keyboard_handle (unsigned char key, int x, int y)
{
    switch (key)
    {
        case ' ':
            ROTATE = !ROTATE;
        break;
        case 'g': case 'G':
            DISPLAY_GRID = !DISPLAY_GRID;
        break;
	case 'b': case 'B':
            DISPLAY_BOX = !DISPLAY_BOX;
        break;
	case '+':
		SEUIL+=0.01;
		cout << "seuil="<<SEUIL<<"\n";
		break;
	case '-':
		SEUIL-=0.01;
		if(SEUIL<0)
			SEUIL=0;
		cout << "seuil="<<SEUIL<<"\n";
		break;
	case 'w' : case 'W' :
		PX_M2 -= 5;
		//if( PX_M2 <0) PX_M2=0;
		cout << "PX_M2=" << PX_M2 <<"\n";
		//M2->Set_Px(PX_M2);
		break;
	case 'x' : case 'X' :
		PX_M2 += 5;
		//if( PX_M2 > MATRIX_SIZE) PX_M2=MATRIX_SIZE;
		cout << "PX_M2=" << PX_M2 <<"\n";
		//M2->Set_Px(PX_M2);
		break;
	case 'p':case 'P':
		The_Matrix->print_line_yz( 160,160);
		break;


    }
}


void keyboard_s (int key, int x, int y)
{
    switch (key)
    {
        case GLUT_KEY_LEFT:
	    rot--;
	    if(rot<0) rot = 359;
        break;
        case GLUT_KEY_RIGHT:
	    rot++;
	    if(rot>=360) rot=0;
        break;
    }
}

int main(int argc ,char ** argv)
{
	glutInit( & argc , argv );
	glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
	glutInitWindowSize( 640,480);
	glutInitWindowPosition(100,100);
	glutCreateWindow("Metaballs 3D");
	glutDisplayFunc(display);
	glutIdleFunc(idle);
	glutReshapeFunc (reshape);
	glutKeyboardFunc (keyboard_handle);
	glutSpecialFunc (keyboard_s);
	init();
	glutMainLoop();

	return 0;
}




